/*
Copyright (C) 2015 Ruslan Kabatsayev <b7.10110111@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FLOAT80_EDIT_H_20151031_
#define FLOAT80_EDIT_H_20151031_

#include "Types.h"
#include <QLineEdit>

namespace ODbgRegisterView {

class Float80Edit : public QLineEdit {
	Q_OBJECT

public:
	explicit Float80Edit(QWidget *parent = nullptr);
	void setValue(edb::value80 input);

public:
	QSize sizeHint() const override;

protected:
	void focusOutEvent(QFocusEvent *e) override;

Q_SIGNALS:
	void defocussed();
};

}

#endif
