/*
Copyright (C) 2006 - 2015 Evan Teran
                          evan.teran@gmail.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "DialogMemoryAccess.h"

namespace DebuggerCorePlugin {

/**
 * @brief DialogMemoryAccess::DialogMemoryAccess
 * @param parent
 * @param f
 */
DialogMemoryAccess::DialogMemoryAccess(QWidget *parent, Qt::WindowFlags f)
	: QDialog(parent, f) {

	ui.setupUi(this);
	adjustSize();
	setFixedSize(width(), height());
}

/**
 * @brief DialogMemoryAccess::warnNextTime
 * @return
 */
bool DialogMemoryAccess::warnNextTime() const {
	return !ui.checkNeverShowAgain->isChecked();
}

}
